#! /usr/bin/env python 
from xtrans_sgmlop import *
"""$Id: extrans.py,v 1.4 2004/06/18 19:41:36 gray Exp $"""

from time import time
import os, string, stat, fphash

import sys
import getopt

def check_for_update(xtransfile, xmlfiles, updatedict, refdate, depxml, depxtrans):
    if depxtrans.has_key(xtransfile):
        xtransdate = max( max(map(os.path.getmtime, depxtrans[xtransfile])), os.path.getmtime(xtransfile) )
    else:
        xtransdate = os.path.getmtime(xtransfile)
    for fi in range(len(xmlfiles)):
        f = xmlfiles[fi] = os.path.abspath(xmlfiles[fi])
        if not updatedict.has_key(f) and (os.path.getmtime(f) > refdate or xtransdate > refdate):
            updatedict[f] = 1
            if depxml.has_key(f):
                for df in depxml[f]:
                    updatedict[df] = 1


extrans_env = {}

def add_define(str):
    s = str.split("=",1)
    extrans_env[s[0]] = s[1]

def usage(code=0):
    print "usage: "+sys.argv[0]+" [-nomaindep][-verbose][-Dvar=value] file.main.xtrans\n" 
    sys.exit(code)	

def main():
    verbose_flag = 0

    st = time();

    try:
	opts, args = getopt.getopt(sys.argv[1:], "hnvD:", ["help", "nomaindep", "verbose", "define"])
    except getopt.GetoptError:
        usage(1)

    for o, a in opts:
	if o in ("-h", "--help"):
		usage()
	if o in ("-n", "--nomaindep"):
	        maindep = 0
	if o in ("-v", "--verbose"):
		verbose_flag = verbose_flag + 1
	if o in ("-D", "--define"):
		add_define(a)

    if len(args) == 1:
        filename = args[0]
    else:
	usage(1)

    # first parse the main.xtrans file
    if verbose_flag:
	print "Reading",filename, "and selecting files to be updated..."
    dirname, basename = os.path.split(filename)
    if dirname:
        os.chdir(dirname)
        if verbose_flag:
		print "changed into", dirname
    filein = open (basename)
    defstart = re.compile("(\w+):\s*(\S+)")
    # last successful run date
    try:
        lastsuccess = os.path.getmtime("."+filename)
    except:
        lastsuccess =  0
    else:
        # we generally prefer to recompile everything when the main file has been modified
        if lastsuccess < os.path.getmtime(filename) and maindep:
            lastsuccess =  0
    # XML files that need to be recompiled
    needupdate = {}
    xmllist = []
    transmap = {}
    mainmap = {}
    targetdir = ""
    targetsubdir = ""
    dbdir = ""
    while 1:
        data = filein.readline()
        if not data:    # end of file
            if xmllist:
                transmap[xtranssource] = xmllist
                check_for_update(xtranssource, xmllist, needupdate, lastsuccess, dependxml, dependxtrans)
            if targetsubdir:
                mainmap[targetdir+'/'+targetsubdir] = transmap
            elif targetdir:
                mainmap[targetdir] = transmap
            else: raise "no 'targetdir' found!"
            break
        if data[0]=='#': continue # skip comments
        m = defstart.match(data)
        if m:
            if xmllist:              # a new statement means the last 'apply' is complete
                transmap[xtranssource] = xmllist
                check_for_update(xtranssource, xmllist, needupdate, lastsuccess, dependxml, dependxtrans)
                xmllist = []
            type = string.lower(m.group(1))
            name = m.group(2)
            if verbose_flag > 1:
		print "found", type, name
            if type == "targetdir": targetdir = name
            elif type == "dbdir":
                dbdir = name
                # open dependency database
                dependxml = fphash.open(dbdir+"/depend_xml.db",'r')
                dependxtrans = fphash.open(dbdir+"/depend_xtrans.db",'r')
            elif type == "targetsubdir":
                if targetsubdir:    # if previous subdir statement, attach collected data to it
                    mainmap[targetdir+'/'+targetsubdir] = transmap
                elif targetdir:
                    mainmap[targetdir] = transmap
                else: raise "Found 'targetsubdir' statement without any targetdir"
                transmap = {}
                targetsubdir = name
            elif type == "apply":
                if not dbdir: raise "No dbdir defined when APPLY statement encountered!"
                xtranssource = os.path.abspath(name)
            else: raise "Unknown type "+type
            continue
        if type == "apply":
            l = string.split(data)
            xmllist.extend(l)

    dependxml.close()
    dependxtrans.close()
    # now process the given files with the appropriate xtrans
    xtrans_objs = []
    parser = XMLParser()
    if verbose_flag:
	print "dbdir is", dbdir
    totfiles = len (needupdate)
    if not totfiles:
    	if verbose_flag:	
        	print("Everything is up to date.")
	sys.exit(0)
    if verbose_flag:	
	print "Parsing (%d files)..." % totfiles
    for targetdir, transmap in mainmap.items():
	if verbose_flag:
		print "  Targetdir:", targetdir
        for xts in transmap.keys():
            files_to_update = []
            for xmlfile in transmap[xts]:
                if needupdate.has_key(xmlfile):
                    files_to_update.append(xmlfile)
            if files_to_update:
                Xtrans = make_Xtrans(xts, dbdir, extrans_env)
		if verbose_flag:
			print "    Applying", xts,"[%d file(s) to update]" % \
                              len(files_to_update)
                for xmlfile in files_to_update:
                    xtrans_inst = Xtrans()
                    parser.register(xtrans_inst)
                    parser.parseFile(xmlfile, targetdir, dbdir)
                    xtrans_objs.append(xtrans_inst)

    if verbose_flag:	
	print "Postprocessing (%s groups)..." % len(xtrans_objs)
    for xtrans_inst in xtrans_objs:
        xtrans_inst.postprocess()

    lastsuccess = time()
    try:
        os.utime("."+filename, (lastsuccess, lastsuccess))
    except:
        flagfile = open("."+filename,'w')
        flagfile.write("Do not edit or overwrite this file ! It is a timestamp for eXtrans.")
    if verbose_flag:	
	print "\nTiming:", lastsuccess - st


if __name__ == '__main__':
    main() 
	