import  string, os, re, sys, fphash, shutil #, tempfile
from time import asctime, localtime, time
#from xml.parsers import sgmlop
import sgmlop
import traceback

"""$Id: xtrans_sgmlop.py,v 1.4 2004/06/18 19:42:31 gray Exp $"""

# OPTIMIZE: look for class-wide declarations (like reg xpressions)


# Base class for XTranslators
class XTransBase:
    """Base class for XTranslators. Hide all stuff specific to the
    underlying parser it relies on (sgmlop for this one).
    That is why it should be derived to be useful (this base is an abstraction only)."""

    generator="eXtrans tools 1.2.1, $Revision: 1.4 $, (C) 2000 Romuald Texier, (C) 2004 Sergey Poznyakoff"
    # regular expression for matching XML file ref
    xmlref = re.compile("[\w/\-\.]+\.xml", re.UNICODE)

    def __init__(self, doctype, ctags, atags, nostrip, tformat):
        """All derived classes must call this base class constructor in order to initialize
        internals, so be carefull if you override it"""

        # "private" attributes
        self.ctags = ctags
        self.atags = atags
        self.nostrip = nostrip
        self.open_session = None

        # "public" attribute (read-only except when mentioned otherwise)
        # see also open_parse_session below
        self.targetformat = tformat
        self.doctype = doctype
        self.date= asctime(localtime(time()))
        self.content = {}
        self.attr = {}

    def open_parse_session(self, filename, targetdir, dbdir):
        """Some initializations occuring just before a new parse session starts
        Used by XML parser to give the xmlsource path and targetdir.
        xmlsource (filename) is needed to resolve relative links. filename
        should be an absolute path. This function stores this absolute path
        in self.xmlsource, sets self.xmlprefix and self.xmldir.
        targetdir is made absolute before being stored in self.targetdir"""

        self.locdbname = os.path.abspath(dbdir+"/locate_"+self.targetformat+".db")
        # dependencies between XML files
        self.depdbname = os.path.abspath(dbdir+"/depend_xml.db")
        self.xmlsource = filename
        self.xmldir, name = os.path.split(self.xmlsource)
        self.xmldir = self.xmldir + '/'
        if name[-4:] == ".xml":
            self.xmlprefix = name[:-4]
        else:
            self.handle_error("file without .xml extension")

        self.targetdir = os.path.abspath(targetdir)
        if self.targetdir[-1] != '/':
            self.targetdir = self.targetdir + '/'

        self.filenames = [] # keep filenames for current session, relative to targetdir
        self.labels = [] # keep labels in order of creation. Labels are not necessarily unique
        self.output = open ("/dev/null", 'w')   # dirty hack
        self.tagstack = []  # used to check well formed XML
        self.locate = fphash.open(self.locdbname, 'c')
        self.depend = fphash.open(self.depdbname, 'c')
        # remove any occurence of this xmlfile in the dependency DB since links may have changed
        for key in self.depend.keys():
            while 1:
                try:
                    self.depend[key].remove(self.xmlsource)
                except: break
        self.open_session = 1
        self.tmpind = 0

    def create_file_name(l, base):
	"""Create new output file name"""
	return "%s_%02d_%s.%s" % (self.xmlprefix, l, base, self.targetformat)


    def set_output(self, label, base):
        """Set new output"""

        self.output.close()
        self.labels.append(label)
        l = len(self.labels)
	filename = self.create_file_name(l, base)
        self.filenames.append(filename)
        # make filename an absolute pathname to avoid ambiguity in DB
        filename = os.path.normpath(self.targetdir + filename)
        #print "Opening", filename, "for output"
        try:
            self.output = open(filename, 'w')
        except:
            raise
        if l == 1:
            if self.locate.has_key(self.xmlsource) and self.locate[self.xmlsource] != filename:
                print "target file",  self.locate[self.xmlsource], "replaced by", filename, "maybe all the serie too"
                #os.remove(self.locate[self.xmlsource])
            self.locate[self.xmlsource]=filename

    def use_tmp_output (self, base, suffix):
        """Use a temporary/auxiliary output. Use it for minor pages only and immediate
         use in main page, not important structures.
         It short-cicuits self.output in the convenience method self.put().
         Return value : the basename (i.e. : path relative to main page)"""

        self.tmpind = self.tmpind + 1
        filename = "%s_aux%02d_%s.%s" % (self.xmlprefix, self.tmpind, base, suffix)
        try:
            self.tmp_output = open(self.targetdir+filename,'w')
        except:
            raise
        return filename

    def close_tmp_output(self):
        """close and delete tmp_output object"""
        self.tmp_output.close()
        del (self.tmp_output)

    def put(self, s):
        """Convenience function that calls self.tmp_output.write(s)
        or self.output.write(s)"""
        try:
            self.tmp_output.write(s)
        except:
            self.output.write(s)

    def is_parent(self, lotag):
        """Is lotag the last opened and not closed tag?"""
        if self.tagstack:
            return lotag == self.tagstack[-1]
        else:
            return None

    def finish_starttag(self, tag, attrs):
        """Called by the parser when a start tag is encountered. It initializes  content or
        attribute entries if tag is in self.ctags or self.atags. Then it calls the method
        start_TAGNAME() if it exists. start_TAGNAME is typically implemented by derived classes."""


        # initialize content dict entry for tag, and set attribute dict entry for tag
        if tag in self.ctags:
            self.content[tag] = ""
        if tag in self.atags:
            self.attr[tag] = attrs
        # check attributes for links
        for attval in attrs.values():
            m = self.xmlref.search(attval)
            if m:
                xmltarget = os.path.normpath(self.xmldir+m.group())
                if self.depend.has_key(xmltarget):
                    self.depend[xmltarget].append(self.xmlsource)
                    # note that we did not check for duplicate entries, because external
                    # applications may be interested in link counts.
                else:
                    self.depend[xmltarget] = [self.xmlsource]
        # action !
        try:
            func = getattr(self, 'start_' + tag)
        except AttributeError:
            self.default_starttag(tag, attrs)
        else:
            func()

        # XML checking: push open tag in
        self.tagstack.append(tag)

    def default_starttag(self, tag, attrs):
        """to be overridden"""
        pass


    def finish_endtag(self, tag):
        """Called by the parser when an end tag is encountered. After valid XML checking,
        it calls the method end_TAGNAME() if it exists.
        end_TAGNAME is typically implemented by derived classes."""

        # XML checking: pop open tag out and compare to close tag
        opentag = self.tagstack.pop()
        if opentag != tag:
            self.handle_error ("Error! Found end tag "+`tag`+" when "+`opentag`+" expected")
        else: # XML is ok, so action !
            try:
                func = getattr(self, 'end_' + tag)
            except AttributeError:
                self.default_endtag(tag)
            else:
                func()
        if self.attr.has_key(tag) and self.attr[tag].has_key("ID"):  # we do it now, in case func() would have called set_output
            refentry =  self.xmlsource + '#' + self.attr[tag]["ID"]
            if self.locate.has_key(refentry) and self.locate[refentry] != self.output.name:
                print "ID",refentry,"shifted from",  self.locate[refentry], "to", self.output.name
            self.locate[refentry] = self.output.name

    def default_endtag(self, tag):
        """to be overridden"""
        pass

    def handle_data(self, data):
        """Called by the parser for data. This method passes data through to the output or
        to the self.content entry for the current open tag if it is in self.ctags."""

        opentag = self.tagstack and self.tagstack[-1]
        if  not opentag or not self.nostrip.has_key(opentag):
            data = string.strip(data)
            if not data: return
        if opentag in self.ctags:
            self.content[opentag] = self.content[opentag] + data
        else:
            try:
                self.tmp_output.write(data)
            except:
                self.output.write(data)


    def handle_special(self, data):
        """Called by the parser for directives.
        Currently used to check the XML file DOCTYPE"""

        tdata = string.split(data);
        if tdata[0] == "DOCTYPE" and tdata[1] != self.doctype:
            self.handle_error("Wrong DOCTYPE! Found "+`tdata[1]`+" when "+`self.doctype`+" expected")

    def close_parse_session(self):
        """Called by the XMLParser when the end of xml file is reached."""

        self.output.close()
        self.locate.close()
        self.depend.close()
        # XML checking: are there open tags left ?
        if self.tagstack:
            self.handle_error("Error ! Open tag(s) left :"+`self.tagstack`)
        # update main DB
        self.open_session = None

    def postprocess(self):
        """Do all post-processing correspondig to the last closed parse session"""

        # Did we parse? Is the last parse session closed?
        if not self.filenames or self.open_session:
            self.handle_error("attempting to postprocess but no previous parse session was closed!")

        # open locdbname just for reading
        self.locate = fphash.open(self.locdbname)

        # we loop through indexes to enable access to next and previous file in list
        for currentfileno in range(len(self.labels)):
            filename = self.targetdir + self.filenames[currentfileno]
            # tempname = tempfile.mktemp()
            tempname = filename+".tmp"
            self.output = open (tempname, 'w')
            # print "postprocessing of", filename
            filein = open (filename)
            while 1:
                data = filein.readline()
                if not data:
                    break
                # first, substitute file references
                data = self.fileref.sub(self.subfileref,data)
                # second, search for macros (lower precedence). Allow for multiple macros on a line
		while 1:
                    m = self.postmacro.search(data)
                    if m:
                    	try:
                        	func = getattr(self, 'post_' + m.group(1))
                        except AttributeError:
                       		print "Warning: undefined macro", m.group(1), "in", filename
                    	else:
                        	self.output.write(data[:m.start()])
                        	func(currentfileno)
				data = data[m.end():]
	                        continue
		    else:
			break
                # third, common data
                self.output.write(data)
            self.output.close()
            filein.close()
            os.rename(tempname, filename) # python 1.5.2 buggy
            # shutil.copy(tempname, filename)
            # os.remove(tempname)
        self.locate.close()

    def subfileref(self, m):
        if m.group(2):   # link (in)to another XML file
            if os.path.isabs(m.group(1)):
                refentry = m.group(1)
            else:
                refentry = os.path.normpath (self.xmldir+m.group(1))
        elif m.group(3): # link within current XML file
            refentry = self.xmlsource + m.group(3)
        else: # we matched [] which is magic
            refentry = self.xmlsource
        try:
            tapath = self.locate[refentry]
            # we want relative paths for portability
            pl = len(os.path.commonprefix((tapath, self.targetdir)))
            retval = "../" * (string.count(self.targetdir[pl:],'/')) + tapath[pl:]
        except:
            print "Error: unresolved reference to", refentry
            retval = "error."+self.targetformat
        return retval


    def handle_error(self, err):
        """Handle all errors. Base implementation just raises err"""

        raise err



# usually implemented by derived classes to do something:
# methods in the form start_TAGNAME(self) or end_TAGNAME(self)
# or post_POSTMACRONAME(self, currentfileno)


# Can be implemented by derived classes:

#    def handle_entityref(self, data):
#        pass

#    def handle_proc(self, name, data):
#        pass

#    def handle_cdata(self, data):
#        print "CData:", data

#    def handle_charref(self, data):
#        pass

#    def handle_comment(self, data):
#        pass


# END OF CLASS  XTransBase


def print_error(filename, lineno, msg):
	print filename+":"+str(lineno)+": "+msg


def exception_handler(inputline, inputname, defname):
        try:
            type, value, tb = sys.exc_info()
 	    msg, (filename, lineno, offset, line) = value
	    if not filename:
			print_error(inputname,inputline+lineno,msg)
#			exc = traceback.format_exception_only(type, value)
#                        print_error(inputname,inputline+lineno,None,exc[1])
#                        print_error(inputname,inputline+lineno,None,exc[2])
	    else:
			print_error(filename,lineno,msg)
        finally:
            tblist = tb = None
	raise

# recursive part of make_Xtrans

def rec_trans_parse(filename, defstart, ctagre, atagre, ctags, atags, nostrip, MyHandler, deplist):
    newdef = ''
    filein = open (filename)
    path = os.path.dirname(filename)
    linecount = 0
    while 1:
        data = filein.readline()
        linecount = linecount+1
        if not data:
            if newdef:
                try:
		    exec(newdef, globals(), MyHandler.__dict__)
                except:
		    exception_handler(startline,filename,newdef)
            break
        if data[0]=='#':
	    continue # skip comments
        m = defstart.match(data)
        if m:
            if newdef:
                try:
                    exec(newdef, globals(), MyHandler.__dict__)
                except:
		    exception_handler(startline,filename,newdef)
            type = string.lower(m.group(1))
            name = m.group(2)
            params = data[m.end():]
            # print "found", type, name
            if type == "starttag":
                if name == "*":
                    newdef = "def default_starttag(xtrans, tag, attrs):\n"
                else:
                    newdef = "def start_"+name+"(xtrans):\n"
                if string.find (params, "nostrip") > -1:
                    nostrip[name]=1
            elif type == "endtag":
                if name == "*":
                    newdef = "def default_endtag(xtrans, tag):\n"
                else:
                    newdef = "def end_"+name+"(xtrans):\n"
            elif type == "postmacro":
                newdef = "def post_"+name+"(xtrans, currentfileno):\n"
            elif type == "global":
                newdef = name+"="
            elif type == "function":
                newdef = "def "+name+"(xtrans,"+params+"):\n"
            elif type == "import":
                exec ("import "+name, globals())
            elif type == "include":
                newpath = os.path.normpath(path+'/'+name)
                deplist.append(newpath)
                rec_trans_parse(newpath, defstart, ctagre, atagre,
	                        ctags, atags, nostrip,MyHandler, deplist)
            elif type == "ppdelim":
                ldel = name
                rdel = string.split(params)[0]
                newdef = """fileref = re.compile("%s(([\w/\-\.]+\.xml)?(#(\w+))?)%s", re.UNICODE)\n""" % (ldel, rdel)
                newdef = newdef + """postmacro = re.compile("%s(\w+)%s", re.UNICODE)\n""" % (ldel, rdel)

            else:
		print_error(filename,linecount,"Unknown type "+type)
		sys.exit(1)
            startline = linecount-1
            continue
        l = ctagre.findall(data)
        if l:
            for tag in l:
                if tag not in ctags:
                    ctags.append (tag)
        l = atagre.findall(data)
        if l:
            for tag in l:
                if tag not in atags:
                    atags.append (tag)
        newdef = newdef + data

# return a translator class derived from XTransBase build upon an abstract description

def make_Xtrans (filename, dbdir, env={}):
    """Returns a translator class build from the abstract description in filename.
    A translator instance is a handler for an XMLParser instance"""

    class MyHandler(XTransBase):
	environ = {}

    # XTrans dependency DB
    depend = fphash.open(dbdir+"/depend_xtrans.db",'c')
    depend[filename] = []
    basename = os.path.basename(filename)
    dec = string.split(basename, '.')
    dtype = dec[0]
    tformat = dec[1]
    ctags = []  # tags to keep content of
    atags = []  # tags to keep attributes of
    nostrip = {}  # marks tags in which leading and trailing white spaces are relevant
    defstart = re.compile("(\w+):\s*(\S+)")
    ctagre = re.compile('xtrans\.content[^"]+"(\w+)"')
    atagre = re.compile('xtrans\.attr[^"]"(\w+)"')

    # print "Generating python class for", dtype, "document type"


    rec_trans_parse(filename, defstart, ctagre, atagre, ctags, atags, nostrip, MyHandler, depend[filename])
    if not depend[filename]:
        del depend[filename]
    if not MyHandler.__dict__.has_key("fileref"):
        # some useful regular expressions
        newdef = """fileref = re.compile("\[(([\w/\-\.]+\.xml)?(#(\w+))?)\]", re.UNICODE)
postmacro = re.compile("\[(\w+)\]", re.UNICODE)\n"""
    else:
        newdef=""

    newdef = newdef + "def __init__(xtrans): XTransBase.__init__(xtrans, %s, %s, %s, %s, %s)" % (`dtype`, `ctags`, `atags`, `nostrip`, `tformat`)
    try: exec(newdef, globals(), MyHandler.__dict__)
    except:
        raise
    newdef = "xtranssource = '%s'" % os.path.abspath(filename)
    try: exec(newdef, globals(), MyHandler.__dict__)
    except:
        raise
    depend.close()
    MyHandler.environ = env
    return MyHandler






# wrapper class for sgmlop.XMLParser
# "Friend" class to XTransBase


class XMLParser:
    """Define the XMLParser to triger XTransBase derived classes. Rely on the underlying parser
    in use (sgmlop for this one)"""

    def __init__(self):
        self.parser = sgmlop.XMLParser()
        self.buffsize =  16384
        self.handler=None

    def register(self, handler):
        """Register a handler of type XTransBase or derived."""

        self.handler = handler
        self.parser.register(handler)

    def parseFile(self, filename, targetdir, dbdir):
        """Parse the given filename"""

        if not self.handler:
            raise "No handler given!"

        self.handler.open_parse_session(filename, targetdir, dbdir)

        fp = open(filename)
        while 1:
            data = fp.read(self.buffsize)
            if not data:
                break
            self.parser.feed(data)
        self.parser.close()
        fp.close()
        self.handler.close_parse_session()


# END OF CLASS XMLParser

